"""Common code for the Kili exporter."""

import json
from pathlib import Path
from typing import Callable, Dict, List

from kili.services.export.exceptions import NotCompatibleInputType, NotCompatibleOptions
from kili.services.export.format.base import AbstractExporter
from kili.services.export.media.video import cut_video
from kili.services.types import Job


class KiliExporter(AbstractExporter):
    """Common code for Kili exporters."""

    ASSETS_DIR_NAME = "assets"

    def _check_arguments_compatibility(self) -> None:
        """Check if the export label format is compatible with the export options."""

    def _check_project_compatibility(self) -> None:
        """Check if the export label format is compatible with the project."""

    def _is_job_compatible(self, job: Job) -> bool:
        """Check if the export label format is compatible with the job."""
        _ = job
        return True  # kili format is compatible with all jobs

    def _save_assets_export(self, assets: List[Dict], output_filename: Path) -> None:
        """Save the assets to a file and return the link to that file."""
        self.logger.info("Exporting to kili format...")

        if self.with_assets:
            if self.project["inputType"] == "VIDEO":
                assets = self._cut_video_assets(assets)

            assets = self._clean_filepaths(assets)

        if self.single_file:
            project_json = json.dumps(assets, sort_keys=True, indent=4)
            self.base_folder.mkdir(parents=True, exist_ok=True)
            with (self.base_folder / "data.json").open("wb") as output_file:
                output_file.write(project_json.encode("utf-8"))
        else:
            labels_folder = self.base_folder / "labels"
            labels_folder.mkdir(parents=True, exist_ok=True)
            for asset in assets:
                external_id = asset["externalId"].replace(" ", "_")
                asset_json = json.dumps(asset, sort_keys=True, indent=4)
                file_path = labels_folder / f"{external_id}.json"
                file_path.parent.mkdir(parents=True, exist_ok=True)
                with file_path.open("wb") as output_file:
                    output_file.write(asset_json.encode("utf-8"))

        self.create_readme_kili_file(self.export_root_folder)

        self.make_archive(self.export_root_folder, output_filename)

        self.logger.warning(output_filename)

    def _clean_filepaths(self, assets: List[Dict]) -> List[Dict]:
        # pylint: disable=line-too-long
        """Remove TemporaryDirectory() prefix from filepaths in "jsonContent" and "content" fields."""
        for asset in assets:
            if Path(asset["content"]).is_file():
                asset["content"] = str(Path(self.ASSETS_DIR_NAME) / Path(asset["content"]).name)

            json_content_list = []
            if isinstance(asset["jsonContent"], list):
                json_content_list = [
                    str(Path(self.ASSETS_DIR_NAME) / Path(filepath).name)
                    for filepath in asset["jsonContent"]
                    if Path(filepath).is_file()
                ]

                asset["jsonContent"] = json_content_list
        return assets

    def _cut_video_assets(self, assets: List[Dict]) -> List[Dict]:
        """Cut video assets into frames."""
        for asset in assets:
            if asset["jsonContent"] == "" and Path(asset["content"]).is_file():
                nbr_frames = len(asset.get("latestLabel", {}).get("jsonResponse", {}))
                if nbr_frames == 0:
                    continue
                leading_zeros = len(str(nbr_frames))
                asset["jsonContent"] = cut_video(
                    video_path=asset["content"],
                    asset=asset,
                    leading_zeros=leading_zeros,
                    output_dir=self.images_folder,
                )
        return assets

    def process_and_save(self, assets: List[Dict], output_filename: Path) -> None:
        """Extract formatted annotations from labels and save the json in the buckets."""
        clean_assets = self.preprocess_assets(assets)

        if self.normalized_coordinates is False:
            clean_assets = [self.convert_to_pixel_coords(asset) for asset in clean_assets]

        return self._save_assets_export(
            clean_assets,
            output_filename,
        )

    @property
    def images_folder(self) -> Path:
        """Export images folder."""
        return self.base_folder / self.ASSETS_DIR_NAME

    def convert_to_pixel_coords(self, asset: Dict) -> Dict:
        """Convert asset JSON response normalized vertices to pixel coordinates."""
        if asset.get("latestLabel", {}):
            self._scale_label_vertices(asset["latestLabel"], asset)

        if asset.get("labels"):
            for label in asset["labels"]:
                self._scale_label_vertices(label, asset)

        return asset

    def _scale_label_vertices(self, label: Dict, asset: Dict) -> None:
        if not label.get("jsonResponse", {}):
            return

        is_label_rotated = (
            label["jsonResponse"]["ROTATION_JOB"]["rotation"] in [90, 270]
            if "ROTATION_JOB" in label["jsonResponse"]
            else False
        )

        if self.project["inputType"] == "PDF":
            self._scale_json_response_vertices(
                label["jsonResponse"],
                asset,
                is_label_rotated,
                _scale_normalized_vertices_pdf_annotation,
            )

        elif self.project["inputType"] == "IMAGE":
            self._scale_json_response_vertices(
                label["jsonResponse"],
                asset,
                is_label_rotated,
                _scale_normalized_vertices_image_video_annotation,
            )

        elif self.project["inputType"] == "VIDEO":
            for frame_resp in label["jsonResponse"].values():
                if frame_resp:
                    self._scale_json_response_vertices(
                        frame_resp,
                        asset,
                        is_label_rotated,
                        _scale_normalized_vertices_image_video_annotation,
                    )

        else:
            raise NotCompatibleInputType(
                f"Labels of input type {self.project['inputType']} cannot be converted to pixel"
                " coordinates."
            )

    def _scale_json_response_vertices(
        self,
        json_resp: Dict,
        asset: Dict,
        is_label_rotated: bool,
        annotation_scaler: Callable[[Dict, Dict, bool], None],
    ) -> None:
        for job_name in json_resp:
            if self._can_scale_vertices_for_job_name(job_name) and json_resp.get(job_name, {}).get(
                "annotations"
            ):
                for ann in json_resp[job_name]["annotations"]:
                    annotation_scaler(ann, asset, is_label_rotated)

    def _can_scale_vertices_for_job_name(self, job_name: str) -> bool:
        return (
            # some old labels might not up to date with the json interface
            job_name in self.project["jsonInterface"]["jobs"]
            and (
                self.project["jsonInterface"]["jobs"][job_name]["mlTask"] == "OBJECT_DETECTION"
                or (
                    self.project["inputType"] == "PDF"
                    and self.project["jsonInterface"]["jobs"][job_name]["mlTask"]
                    == "NAMED_ENTITIES_RECOGNITION"  # PDF NER jobs have vertices
                )
            )
        )


def _scale_vertex(vertex: Dict, width: int, height: int) -> Dict:
    return {"x": vertex["x"] * width, "y": vertex["y"] * height}


def _scale_all_vertices(object_, width: int, height: int):
    if isinstance(object_, List):
        return [_scale_all_vertices(obj, width=width, height=height) for obj in object_]

    if isinstance(object_, Dict):
        if sorted(object_.keys()) == ["x", "y"]:
            return _scale_vertex(object_, width=width, height=height)
        return {
            key: _scale_all_vertices(value, width=width, height=height)
            for key, value in object_.items()
        }

    return object_


def _scale_normalized_vertices_pdf_annotation(
    annotation: Dict, asset: Dict, is_label_rotated: bool = False
) -> None:
    """Scale normalized vertices of a PDF annotation.

    PDF annotations are different from image annotations because the asset width/height can vary.

    PDF only have BBox detection, so we only scale the boundingPoly and polys keys.
    """
    if is_label_rotated:
        raise NotCompatibleOptions("PDF labels cannot be rotated")

    if "annotations" in annotation:
        # pdf annotations have two layers of "annotations"
        # https://docs.kili-technology.com/reference/export-object-entity-detection-and-relation#ner-in-pdfs
        for ann in annotation["annotations"]:
            _scale_normalized_vertices_pdf_annotation(ann, asset)

    # an annotation has three keys:
    # - pageNumberArray: list of page numbers
    # - polys: list of polygons
    # - boundingPoly: list of bounding polygons
    # each polygon is a dict with a key "normalizedVertices" that is a list of vertices
    if "polys" in annotation and "boundingPoly" in annotation:
        try:
            page_number_to_dimensions = {
                page_resolution["pageNumber"]: {
                    "width": page_resolution["width"],
                    "height": page_resolution["height"],
                }
                for page_resolution in asset["pageResolutions"]
            }
        except (KeyError, TypeError) as err:
            raise NotCompatibleOptions(
                "PDF labels export with absolute coordinates require `pageResolutions` in the"
                " asset. Please use `kili.update_properties_in_assets(page_resolutions_array=...)`"
                " to update the page resolutions of your asset.`"
            ) from err

        for key in ("polys", "boundingPoly"):
            annotation[key] = [
                {
                    **value,  # keep the original normalizedVertices
                    "vertices": _scale_all_vertices(
                        value["normalizedVertices"],
                        width=page_number_to_dimensions[page_number]["width"],
                        height=page_number_to_dimensions[page_number]["height"],
                    ),
                }
                for value, page_number in zip(annotation[key], annotation["pageNumberArray"])
            ]


def _scale_normalized_vertices_image_video_annotation(
    annotation: Dict, asset: Dict, is_label_rotated: bool
) -> None:
    """Scale normalized vertices of an image/video object detection annotation."""
    if "resolution" not in asset or asset["resolution"] is None:
        raise NotCompatibleOptions(
            "Image and video labels export with absolute coordinates require `resolution` in the"
            " asset. Please use `kili.update_properties_in_assets(resolution_array=...)` to update"
            " the resolution of your asset.`"
        )

    width = asset["resolution"]["width"] if not is_label_rotated else asset["resolution"]["height"]
    height = asset["resolution"]["height"] if not is_label_rotated else asset["resolution"]["width"]

    # bbox, segmentation, polygons
    if "boundingPoly" in annotation:
        annotation["boundingPoly"] = [
            {
                **norm_vertices_dict,  # keep the original normalizedVertices
                "vertices": _scale_all_vertices(
                    norm_vertices_dict["normalizedVertices"], width=width, height=height
                ),
            }
            for norm_vertices_dict in annotation["boundingPoly"]
        ]

    # point jobs
    if "point" in annotation:
        annotation["pointPixels"] = _scale_all_vertices(
            annotation["point"], width=width, height=height
        )

    # line, vector jobs
    if "polyline" in annotation:
        annotation["polylinePixels"] = _scale_all_vertices(
            annotation["polyline"], width=width, height=height
        )

    # pose estimation jobs
    if "points" in annotation:
        for point_dict in annotation["points"]:
            if "point" in point_dict:
                point_dict["pointPixels"] = _scale_all_vertices(
                    point_dict["point"], width=width, height=height
                )



"""Types for the Cloud storage related Kili API gateway functions."""

from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional

from kili.domain.cloud_storage import (
    DataIntegrationId,
    DataIntegrationPlatform,
    DataIntegrationStatus,
)
from kili.domain.organization import OrganizationId
from kili.domain.project import ProjectId


@dataclass
class AddDataConnectionKiliAPIGatewayInput:
    """Add data connection input data for Kili API Gateway."""

    is_checking: bool
    integration_id: DataIntegrationId
    last_checked: datetime
    project_id: ProjectId
    selected_folders: Optional[List[str]]


@dataclass
class DataConnectionComputeDifferencesKiliAPIGatewayInput:
    """Data connection compute differences input data for Kili API Gateway."""

    blob_paths: List[str]
    warnings: List[str]
    content_types: List[str]


@dataclass
class DataIntegrationData:
    """Data integration input data."""

    allowed_paths: Optional[List[str]]
    allowed_projects: Optional[List[str]]
    aws_access_point_arn: Optional[str]
    aws_role_arn: Optional[str]
    aws_role_external_id: Optional[str]
    azure_connection_url: Optional[str]
    azure_is_using_service_credentials: Optional[bool]
    azure_sas_token: Optional[str]
    azure_tenant_id: Optional[str]
    gcp_bucket_name: Optional[str]
    include_root_files: Optional[str]
    internal_processing_authorized: Optional[str]
    name: Optional[str]
    organization_id: Optional[OrganizationId]
    platform: Optional[DataIntegrationPlatform]
    status: Optional[DataIntegrationStatus]
    s3_access_key: Optional[str]
    s3_bucket_name: Optional[str]
    s3_endpoint: Optional[str]
    s3_region: Optional[str]
    s3_secret_key: Optional[str]
    s3_session_token: Optional[str]



"""Project mutations."""

from typing import Dict, Literal, Optional

from typeguard import typechecked

from kili.entrypoints.base import BaseOperationEntrypointMixin
from kili.entrypoints.mutations.exceptions import MutationError
from kili.services.copy_project import ProjectCopier
from kili.utils.logcontext import for_all_methods, log_call

from .queries import (
    GQL_APPEND_TO_ROLES,
    GQL_DELETE_FROM_ROLES,
    GQL_PROJECT_DELETE_ASYNCHRONOUSLY,
    GQL_PROJECT_UPDATE_ANONYMIZATION,
    GQL_UPDATE_PROPERTIES_IN_PROJECT,
    GQL_UPDATE_PROPERTIES_IN_ROLE,
)


@for_all_methods(log_call, exclude=["__init__"])
class MutationsProject(BaseOperationEntrypointMixin):
    """Set of Project mutations."""

    @typechecked
    def append_to_roles(
        self,
        project_id: str,
        user_email: str,
        role: Literal["ADMIN", "TEAM_MANAGER", "REVIEWER", "LABELER"] = "LABELER",
    ) -> Dict:
        """Add a user to a project.

        !!! info
            If the user does not exist in your organization, he/she is invited and added
                both to your organization and project. This function can also be used to change
                the role of the user in the project.

        Args:
            project_id: Identifier of the project
            user_email: The email of the user.
                This email is used as the unique identifier of the user.
            role: The role of the user.

        Returns:
            A dictionary with the project user information.


        Examples:
            >>> kili.append_to_roles(project_id=project_id, user_email='john@doe.com')
        """
        variables = {
            "data": {"role": role, "userEmail": user_email},
            "where": {"id": project_id},
        }
        result = self.graphql_client.execute(GQL_APPEND_TO_ROLES, variables)

        project_data = self.format_result("data", result)
        for project_user in project_data["roles"]:
            if project_user["user"]["email"] == user_email.lower() and project_user["role"] == role:
                return project_user

        raise MutationError(
            f"Failed to mutate user {user_email} to role {role} for project {project_id}."
        )

    @typechecked
    def update_properties_in_role(
        self, role_id: str, project_id: str, user_id: str, role: str
    ) -> Dict:
        """Update properties of a role.

        !!! info
            To be able to change someone's role, you must be either of:

            - an admin of the project
            - a team manager of the project
            - an admin of the organization

        Args:
            role_id: Role identifier of the user. E.g. : 'to-be-deactivated'
            project_id: Identifier of the project
            user_id: The email or identifier of the user with updated role
            role: The new role.
                Possible choices are: `ADMIN`, `TEAM_MANAGER`, `REVIEWER`, `LABELER`

        Returns:
            A dictionary with the project user information.
        """
        variables = {
            "roleID": role_id,
            "projectID": project_id,
            "userID": user_id,
            "role": role,
        }
        result = self.graphql_client.execute(GQL_UPDATE_PROPERTIES_IN_ROLE, variables)
        return self.format_result("data", result)

    @typechecked
    def delete_from_roles(self, role_id: str) -> Dict[Literal["id"], str]:
        """Delete users by their role_id.

        Args:
            role_id: Identifier of the project user (not the ID of the user)

        Returns:
            A dict with the project id.
        """
        variables = {"where": {"id": role_id}}
        result = self.graphql_client.execute(GQL_DELETE_FROM_ROLES, variables)
        return self.format_result("data", result)

    @typechecked
    def delete_project(self, project_id: str) -> str:
        """Delete a project permanently.

        Args:
            project_id: Identifier of the project

        Returns:
            A string with the deleted project id.
        """
        variables = {"where": {"id": project_id}}
        result = self.graphql_client.execute(GQL_PROJECT_DELETE_ASYNCHRONOUSLY, variables)
        return self.format_result("data", result)

    @typechecked
    def archive_project(self, project_id: str) -> Dict[Literal["id"], str]:
        """Archive a project.

        Args:
            project_id: Identifier of the project.

        Returns:
            A dict with the id of the project.
        """
        variables = {
            "projectID": project_id,
            "archived": True,
        }

        result = self.graphql_client.execute(GQL_UPDATE_PROPERTIES_IN_PROJECT, variables)

        return self.format_result("data", result)

    @typechecked
    def unarchive_project(self, project_id: str) -> Dict[Literal["id"], str]:
        """Unarchive a project.

        Args:
            project_id: Identifier of the project

        Returns:
            A dict with the id of the project.
        """
        variables = {
            "projectID": project_id,
            "archived": False,
        }

        result = self.graphql_client.execute(GQL_UPDATE_PROPERTIES_IN_PROJECT, variables)
        return self.format_result("data", result)

    @typechecked
    def copy_project(  # pylint: disable=too-many-arguments
        self,
        from_project_id: str,
        title: Optional[str] = None,
        description: Optional[str] = None,
        copy_json_interface: bool = True,
        copy_quality_settings: bool = True,
        copy_members: bool = True,
        copy_assets: bool = False,
        copy_labels: bool = False,
        disable_tqdm: Optional[bool] = None,
    ) -> str:
        """Create new project from an existing project.

        Args:
            from_project_id: Project ID to copy from.
            title: Title for the new project. Defaults to source project
                title if `None` is provided.
            description: Description for the new project. Defaults to empty string
                if `None` is provided.
            copy_json_interface: Include json interface in the copy.
            copy_quality_settings: Include quality settings in the copy.
            copy_members: Include members in the copy.
            copy_assets: Include assets in the copy.
            copy_labels: Include labels in the copy.
            disable_tqdm: Disable tqdm progress bars.

        Returns:
            The created project ID.

        Examples:
            >>> kili.copy_project(from_project_id="clbqn56b331234567890l41c0")
        """
        return ProjectCopier(self).copy_project(  # pyright: ignore[reportGeneralTypeIssues]
            from_project_id,
            title,
            description,
            copy_json_interface,
            copy_quality_settings,
            copy_members,
            copy_assets,
            copy_labels,
            disable_tqdm,
        )

    @typechecked
    def update_project_anonymization(
        self, project_id: str, should_anonymize: bool = True
    ) -> Dict[Literal["id"], str]:
        """Anonymize the project for the labelers and reviewers.

        !!! info
            Compatible with versions of the Kili app >= 2.135.0

        Args:
            project_id: Identifier of the project
            should_anonymize: The value to be applied. Defaults to `True`.

        Returns:
            A dict with the id of the project which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.update_project_anonymization(project_id=project_id)
            >>> kili.update_project_anonymization(project_id=project_id, should_anonymize=False)
        """
        variables = {
            "input": {
                "id": project_id,
                "shouldAnonymize": should_anonymize,
            }
        }

        result = self.graphql_client.execute(GQL_PROJECT_UPDATE_ANONYMIZATION, variables)
        return self.format_result("data", result)



# pylint: disable=too-many-arguments
"""Client presentation methods for issues."""

from itertools import repeat
from typing import Any, Dict, Generator, Iterable, List, Literal, Optional, overload

from typeguard import typechecked

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.domain.issue import IssueFilters, IssueId, IssueStatus, IssueType
from kili.domain.label import LabelId
from kili.domain.project import ProjectId
from kili.domain.types import ListOrTuple
from kili.presentation.client.helpers.common_validators import (
    assert_all_arrays_have_same_size,
    disable_tqdm_if_as_generator,
)
from kili.use_cases.issue import IssueUseCases
from kili.use_cases.issue.types import IssueToCreateUseCaseInput
from kili.utils.logcontext import for_all_methods, log_call

from .base import BaseClientMethods


@for_all_methods(log_call, exclude=["__init__"])
class IssueClientMethods(BaseClientMethods):
    """Methods attached to the Kili client, to run actions on issues."""

    @typechecked
    def create_issues(
        self,
        project_id: str,
        label_id_array: List[str],
        object_mid_array: Optional[List[Optional[str]]] = None,
        text_array: Optional[List[Optional[str]]] = None,
    ) -> List[Dict[Literal["id"], str]]:
        """Create an issue.

        Args:
            project_id: Id of the project.
            label_id_array: List of Ids of the labels to add an issue to.
            object_mid_array: List of mids of the objects in the labels to associate the issues to.
            text_array: List of texts to associate to the issues.

        Returns:
            A list of dictionaries with the `id` key of the created issues.
        """
        assert_all_arrays_have_same_size([label_id_array, object_mid_array, text_array])
        issues = [
            IssueToCreateUseCaseInput(label_id=LabelId(label_id), object_mid=object_mid, text=text)
            for (label_id, object_mid, text) in zip(
                label_id_array,
                object_mid_array or repeat(None),
                text_array or repeat(None),
            )
        ]
        issue_service = IssueUseCases(self.kili_api_gateway)
        issue_ids = issue_service.create_issues(project_id=ProjectId(project_id), issues=issues)
        return [{"id": issue_id} for issue_id in issue_ids]

    @typechecked
    def count_issues(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_id_in: Optional[List[str]] = None,
        issue_type: Optional[IssueType] = None,
        status: Optional[IssueStatus] = None,
    ) -> int:
        """Count and return the number of issues with the given constraints.

        Args:
            project_id: Project ID the issue belongs to.
            asset_id: Asset id whose returned issues are associated to.
            asset_id_in: List of asset ids whose returned issues are associated to.
            issue_type: Type of the issue to return. An issue object both
                represents issues and questions in the app.
            status: Status of the issues to return.

        Returns:
            The number of issues that match the given constraints.
        """
        if asset_id and asset_id_in:
            raise ValueError(
                "You cannot provide both `asset_id` and `asset_id_in` at the same time."
            )
        return IssueUseCases(self.kili_api_gateway).count_issues(
            IssueFilters(
                project_id=ProjectId(project_id),
                asset_id=asset_id,
                asset_id_in=asset_id_in,
                issue_type=issue_type,
                status=status,
            )
        )

    @overload
    def issues(
        self,
        project_id: str,
        fields: ListOrTuple[str] = (
            "id",
            "createdAt",
            "status",
            "type",
            "assetId",
        ),
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        asset_id: Optional[str] = None,
        asset_id_in: Optional[List[str]] = None,
        issue_type: Optional[IssueType] = None,
        status: Optional[IssueStatus] = None,
        *,
        as_generator: Literal[True],
    ) -> Generator[Dict, None, None]:
        ...

    @overload
    def issues(
        self,
        project_id: str,
        fields: ListOrTuple[str] = (
            "id",
            "createdAt",
            "status",
            "type",
            "assetId",
        ),
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        asset_id: Optional[str] = None,
        asset_id_in: Optional[List[str]] = None,
        issue_type: Optional[IssueType] = None,
        status: Optional[IssueStatus] = None,
        *,
        as_generator: Literal[False] = False,
    ) -> List[Dict]:
        ...

    @typechecked
    def issues(
        self,
        project_id: str,
        fields: ListOrTuple[str] = (
            "id",
            "createdAt",
            "status",
            "type",
            "assetId",
        ),
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        asset_id: Optional[str] = None,
        asset_id_in: Optional[List[str]] = None,
        issue_type: Optional[IssueType] = None,
        status: Optional[IssueStatus] = None,
        *,
        as_generator: bool = False,
    ) -> Iterable[Dict]:
        # pylint: disable=line-too-long
        """Get a generator or a list of issues that match a set of criteria.

        !!! Info "Issues or Questions"
            An `Issue` object both represent an issue and a question in the app.
            To create them, two different methods are provided: `create_issues` and `create_questions`.
            However to query issues and questions, we currently provide this unique method that retrieves both of them.

        Args:
            project_id: Project ID the issue belongs to.
            asset_id: Id of the asset whose returned issues are associated to.
            asset_id_in: List of Ids of assets whose returned issues are associated to.
            issue_type: Type of the issue to return. An issue object both represents issues and questions in the app.
            status: Status of the issues to return.
            fields: All the fields to request among the possible fields for the assets.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#issue) for all possible fields.
            first: Maximum number of issues to return.
            skip: Number of issues to skip (they are ordered by their date of creation, first to last).
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the issues is returned.

        Returns:
            An iterable of issues objects represented as `dict`.

        Examples:
            >>> kili.issues(project_id=project_id, fields=['author.email']) # List all issues of a project and their authors
        """
        if asset_id and asset_id_in:
            raise ValueError(
                "You cannot provide both `asset_id` and `asset_id_in` at the same time."
            )

        disable_tqdm = disable_tqdm_if_as_generator(as_generator, disable_tqdm)
        options = QueryOptions(disable_tqdm=disable_tqdm, first=first, skip=skip)
        issues_gen = IssueUseCases(self.kili_api_gateway).list_issues(
            IssueFilters(
                project_id=ProjectId(project_id),
                asset_id=asset_id,
                asset_id_in=asset_id_in,
                issue_type=issue_type,
                status=status,
            ),
            fields=fields,
            options=options,
        )
        if as_generator:
            return issues_gen
        return list(issues_gen)

    def update_issue_status(self, issue_id: str, status: IssueStatus) -> Dict[str, Any]:
        """Update the status of an issue.

        Args:
            issue_id: Identifier of the issue.
            status: New status of the issue. "OPEN" or "SOLVED"

        Returns:
            A dict with the changed properties which indicates if the mutation was successful,
                else an error message.
        """
        return IssueUseCases(self.kili_api_gateway).update_issue_status(IssueId(issue_id), status)



"""Annotation domain."""

from typing import List, Literal, NewType, Optional, TypedDict, Union

from .label import LabelId
from .ontology import JobName

AnnotationId = NewType("AnnotationId", str)
AnnotationValueId = NewType("AnnotationValueId", str)
KeyAnnotationId = NewType("KeyAnnotationId", str)


class Vertice(TypedDict):
    """Vertice."""

    x: float
    y: float


class ObjectDetectionAnnotationValue(TypedDict):
    """Object detection annotation value."""

    vertices: List[List[List[Vertice]]]


class ClassificationAnnotationValue(TypedDict):
    """Classification annotation value."""

    categories: List[str]


class ClassificationAnnotation(TypedDict):
    """Classification annotation."""

    # pylint: disable=unused-private-member
    __typename: Literal["ClassificationAnnotation"]
    id: AnnotationId
    labelId: LabelId
    job: JobName
    path: List[List[str]]
    annotationValue: ClassificationAnnotationValue


class RankingOrderValue(TypedDict):
    """Ranking order value."""

    rank: int
    elements: List[str]


class RankingAnnotationValue(TypedDict):
    """Ranking annotation value."""

    orders: List[RankingOrderValue]


class RankingAnnotation(TypedDict):
    """Ranking annotation."""

    # pylint: disable=unused-private-member
    __typename: Literal["RankingAnnotation"]
    id: AnnotationId
    labelId: LabelId
    job: JobName
    path: List[List[str]]
    annotationValue: RankingAnnotationValue


class TranscriptionAnnotationValue(TypedDict):
    """Transcription annotation value."""

    text: str


class TranscriptionAnnotation(TypedDict):
    """Transcription annotation."""

    # pylint: disable=unused-private-member
    __typename: Literal["TranscriptionAnnotation"]
    id: AnnotationId
    labelId: LabelId
    job: JobName
    path: List[List[str]]
    annotationValue: TranscriptionAnnotationValue


class Annotation(TypedDict):
    """Annotation."""

    id: AnnotationId
    labelId: LabelId
    job: JobName
    path: List[List[str]]


class FrameInterval(TypedDict):
    """Frame interval."""

    start: int
    end: int


class VideoObjectDetectionKeyAnnotation(TypedDict):
    """Video object detection key annotation."""

    id: KeyAnnotationId
    frame: int
    annotationValue: ObjectDetectionAnnotationValue


class VideoClassificationKeyAnnotation(TypedDict):
    """Video classification key annotation."""

    id: KeyAnnotationId
    frame: int
    annotationValue: ClassificationAnnotationValue


class VideoTranscriptionKeyAnnotation(TypedDict):
    """Video transcription key annotation."""

    id: KeyAnnotationId
    frame: int
    annotationValue: TranscriptionAnnotationValue


class VideoObjectDetectionAnnotation(TypedDict):
    """Video object detection annotation."""

    # pylint: disable=unused-private-member
    __typename: Literal["VideoObjectDetectionAnnotation"]
    id: AnnotationId
    labelId: LabelId
    job: JobName
    path: List[List[str]]
    frames: List[FrameInterval]
    keyAnnotations: List[VideoObjectDetectionKeyAnnotation]
    name: Optional[str]
    mid: str
    category: str


class VideoClassificationAnnotation(TypedDict):
    """Video classification annotation."""

    # pylint: disable=unused-private-member
    __typename: Literal["VideoClassificationAnnotation"]
    id: AnnotationId
    labelId: LabelId
    job: JobName
    path: List[List[str]]
    frames: List[FrameInterval]
    keyAnnotations: List[VideoClassificationKeyAnnotation]


class VideoTranscriptionAnnotation(TypedDict):
    """Video transcription annotation."""

    # pylint: disable=unused-private-member
    __typename: Literal["VideoTranscriptionAnnotation"]
    id: AnnotationId
    labelId: LabelId
    job: JobName
    path: List[List[str]]
    frames: List[FrameInterval]
    keyAnnotations: List[VideoTranscriptionKeyAnnotation]


VideoAnnotation = Union[
    VideoObjectDetectionAnnotation,
    VideoClassificationAnnotation,
    VideoTranscriptionAnnotation,
]

ClassicAnnotation = Union[
    ClassificationAnnotation,
    RankingAnnotation,
    TranscriptionAnnotation,
]

Annotation = Union[
    ClassificationAnnotation,
    RankingAnnotation,
    TranscriptionAnnotation,
    VideoAnnotation,
]



"""Mappers for Organization API calls."""

import json
from typing import Dict

from kili.domain.organization import (
    OrganizationFilters,
    OrganizationId,
    OrganizationMetricsFilters,
    OrganizationToCreateInput,
    OrganizationToUpdateInput,
)


def map_create_organization_data(data: OrganizationToCreateInput) -> Dict:
    """Build the GraphQL OrganizationData variable to be sent in an operation."""
    return {
        "name": data.name,
        "address": data.address,
        "city": data.city,
        "country": data.country,
        "zipCode": data.zip_code,
    }


def map_list_organizations_where(filters: OrganizationFilters) -> Dict:
    """Build the GraphQL OrganizationWhere variable to be sent in an operation."""
    return {
        "id": filters.organization_id,
        "user": {
            "email": filters.email,
        },
    }


def map_update_organization_where(organization_id: OrganizationId) -> Dict:
    """Build the GraphQL OrganizationData variable to be sent in an operation."""
    return {"id": organization_id}


def map_update_organization_data(organization_data: OrganizationToUpdateInput) -> Dict:
    """Build the GraphQL OrganizationData variable to be sent in an operation."""
    license_str = None if not organization_data.license else json.dumps(organization_data.license)
    data = {}
    if organization_data.name is not None:
        data["name"] = organization_data.name
    if license_str is not None:
        data["license"] = license_str
    return data


def map_organization_metrics_where(filters: OrganizationMetricsFilters) -> Dict:
    """Build the GraphQL OrganizationMetricsWhere variable to be sent in an operation."""
    date_string_fmt = "%Y-%m-%dT%H:%M:%SZ"
    return {
        "organizationId": filters.id,
        "startDate": filters.start_datetime.strftime(date_string_fmt),
        "endDate": filters.end_datetime.strftime(date_string_fmt),
    }



"""GraphQL Cloud Storage operations."""


def get_list_data_integrations_query(fragment: str) -> str:
    """Return the GraphQL query to list data integrations."""
    return f"""
        query dataIntegrations($where: DataIntegrationWhere!, $first: PageSize!, $skip: Int!) {{
          data: dataIntegrations(where: $where, skip: $skip, first: $first) {{
            {fragment}
          }}
        }}
        """


def get_list_data_connections_query(fragment: str) -> str:
    """Return the GraphQL query to list data connections."""
    return f"""
        query dataConnections($where: DataConnectionsWhere!, $first: PageSize!, $skip: Int!) {{
          data: dataConnections(where: $where, first: $first, skip: $skip) {{
            {fragment}
          }}
        }}
        """


def get_data_connection_query(fragment: str) -> str:
    """Return the GraphQL query to get a data connection."""
    return f"""
        query dataConnection($where: DataConnectionIdWhere!) {{
          data: dataConnection(where: $where) {{
            {fragment}
          }}
        }}
        """


def get_add_data_connection_mutation(fragment: str) -> str:
    """Return the GraphQL mutation to add a data connection."""
    return f"""
    mutation addDataConnection($data: DataConnectionInput!) {{
      data: addDataConnection(data: $data) {{
        {fragment}
      }}
    }}
    """


def get_compute_data_connection_differences_mutation(fragment: str) -> str:
    """Return the GraphQL mutation to compute data connection differences."""
    return f"""
    mutation computeDifferences(
        $where: DataConnectionIdWhere!,
        $data: DataConnectionComputeDifferencesPayload
      ) {{
          data: computeDifferences(where: $where, data: $data) {{
              {fragment}
          }}
      }}
    """


def get_validate_data_connection_differences_mutation(fragment: str) -> str:
    """Return the GraphQL mutation to validate data connection differences."""
    return f"""
    mutation validateDataDifferences(
      $where: ValidateDataDifferencesWhere!,
      $processingParameters: String) {{
      data: validateDataDifferences(
        where: $where,
        processingParameters: $processingParameters
      ) {{
        {fragment}
      }}
    }}
    """


GQL_COUNT_DATA_INTEGRATIONS = """
    query countDataIntegrations($where: DataIntegrationWhere!) {
      data: countDataIntegrations(where: $where)
    }
    """


def get_update_integration_mutation(fragment: str) -> str:
    """Return the GraphQL mutation to update a data integrations."""
    return f"""
    mutation UpdatePropertiesInDataIntegration(
      $data: DataIntegrationData!,
      $where: DataIntegrationWhere!) {{
      data: updatePropertiesInDataIntegration(data: $data, where: $where) {{
        {fragment}
      }}
    }}
    """


def get_create_integration_mutation(fragment: str) -> str:
    """Return the GraphQL mutation to create a data integration."""
    return f"""
        mutation CreateDataIntegration($data: DataIntegrationData!) {{
          data: createDataIntegration(data: $data) {{
            {fragment}
          }}
        }}
        """


GQL_DELETE_DATA_INTEGRATION = """
    mutation DeleteDataIntegration($where: DataIntegrationWhere!) {
            data: deleteDataIntegration(where: $where)
        }
    """
